from pathlib import Path

from una import config, sync
from una.types import Diff, Include, OrgImports, Style


def test_int_dep_to_pyproject_package():
    ns = "unit_test"
    int_dep = "greet"
    root = "../../libs"
    expected_proj = Include(src=f"{root}/{ns}/{int_dep}", dst=f"{ns}/{int_dep}")
    style = Style.modules
    res_proj = sync.to_package(ns, int_dep, root, style)
    assert res_proj == expected_proj


def test_int_deps_to_pyproject_packages():
    ns = "unit_test"
    app_name = "hello"
    lib_name = "world"
    expected = [
        Include(src=f"../../apps/{app_name}/{ns}/{app_name}", dst=f"{ns}/{app_name}"),
        Include(src=f"../../libs/{lib_name}/{ns}/{lib_name}", dst=f"{ns}/{lib_name}"),
    ]
    diff = Diff(
        name="unit-test",
        path=Path.cwd(),
        apps={app_name},
        libs={lib_name},
        int_dep_imports=OrgImports(),
    )
    res = sync.to_packages(ns, diff)
    assert res == expected


packages = [
    Include(src="apps/hello/first", dst="hello/first"),
    Include(src="libs/hello/second", dst="hello/second"),
    Include(src="libs/hello/third", dst="hello/third"),
]
expected_hatch_packages = {
    "apps/hello/first": "hello/first",
    "libs/hello/second": "hello/second",
    "libs/hello/third": "hello/third",
}

BASE_PYPROJECT = """
[project]
name = ""
version = ""
description = ""
authors = []
dependencies = []
readme = ""
requires-python = ""
"""


def test_generate_updated_hatch_project_with_existing_una_sections():
    pyproj = f"""
{BASE_PYPROJECT}
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"
[tool.hatch.build]
[tool.una.libs]
"apps/hello/first" = "hello/first"
"""
    conf = config.load_conf_from_str(pyproj)
    updated = str(sync.generate_updated_project(conf, packages[1:]))
    res = config.load_conf_from_str(updated).tool.una.libs
    assert res == expected_hatch_packages


def test_generate_updated_hatch_project_with_missing_int_dep_config():
    pyproj = f"""
{BASE_PYPROJECT}
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"
[tool.hatch.build]
    """
    conf = config.load_conf_from_str(pyproj)
    updated = str(sync.generate_updated_project(conf, packages))
    res = config.load_conf_from_str(updated).tool.una.libs
    assert res == expected_hatch_packages



import difflib
from functools import reduce
from operator import itemgetter
from pathlib import Path

from rich import box, markup
from rich.console import Console
from rich.padding import Padding
from rich.table import Table

from una import config, defaults, distributions, files, parse, stdlib
from una.types import Imports, Options, OrgImports, Proj, Style


def external_deps_from_all(
    root: Path,
    ns: str,
    projects: list[Proj],
    options: Options,
) -> set[bool]:
    imports = {p.name: get_third_party_imports(root, ns, p) for p in projects}
    flattened = reduce(flatten_imports, imports.values(), OrgImports())
    print_libs_summary()
    print_libs_in_int_deps(flattened)
    return {missing_libs(p, imports, options) for p in projects}


def missing_libs(project: Proj, imports: dict[str, OrgImports], options: Options) -> bool:
    name = project.name
    deps = project.ext_deps
    int_dep_imports = imports[name]
    libs = distributions.known_aliases_and_sub_dependencies(deps, options.alias)
    return print_missing_installed_libs(
        int_dep_imports,
        libs,
        name,
    )


def flatten_imports(acc: OrgImports, item: OrgImports) -> OrgImports:
    apps_i = item.apps
    libs_i = item.libs
    return OrgImports(
        apps={**acc.apps, **apps_i},
        libs={**acc.libs, **libs_i},
    )


def extract_top_ns_from_imports(imports: set[str]) -> set[str]:
    return {imp.split(".")[0] for imp in imports}


def extract_ext_dep_imports(all_imports: Imports, top_ns: str) -> Imports:
    std_libs = stdlib.get_stdlibs()
    top_level_imports = {k: extract_top_ns_from_imports(v) for k, v in all_imports.items()}
    to_exclude = std_libs.union({top_ns})
    with_third_party = {k: v - to_exclude for k, v in top_level_imports.items()}
    return {k: v for k, v in with_third_party.items() if v}


def _get_third_party_imports(root: Path, paths: set[Path]) -> Imports:
    top_ns = config.get_ns()
    all_imports = parse.fetch_all_imports(paths)
    return extract_ext_dep_imports(all_imports, top_ns)


def get_third_party_imports(root: Path, ns: str, project: Proj) -> OrgImports:
    apps_d = {b for b in project.int_deps.apps}
    libs_d = {c for c in project.int_deps.libs}
    apps_paths = files.collect_apps_paths(root, ns, apps_d)
    libs_paths = files.collect_libs_paths(root, ns, libs_d)
    apps_imports = _get_third_party_imports(root, apps_paths)
    libs_imports = _get_third_party_imports(root, libs_paths)
    return OrgImports(apps=apps_imports, libs=libs_imports)


def flatten(imports: dict[str, set[str]]) -> set[str]:
    res: set[str] = set().union(*imports.values())
    return res


def calculate_diff(imports: OrgImports, deps: set[str], include_libs: bool) -> set[str]:
    apps_imports = flatten(imports.apps)
    # workspace style doesn't require libs imports to be declared in the app deps
    libs_imports: set[str] = flatten(imports.libs) if include_libs else set()
    flat: set[str] = set().union(apps_imports, libs_imports)
    unknown_imports = flat.difference(deps)
    cutoff = 0.6

    unknowns = {str.lower(u) for u in unknown_imports}
    deps_norm = {str.lower(d).replace("-", "_") for d in deps}
    filtered = {u for u in unknowns if not difflib.get_close_matches(u, deps_norm, cutoff=cutoff)}
    return filtered


def print_libs_summary() -> None:
    console = Console(theme=defaults.una_theme)
    console.print(Padding("[data]Libraries in apps and libs[/]", (1, 0, 0, 0)))


def print_libs_in_int_deps(int_dep_imports: OrgImports) -> None:
    apps_imports = flatten(int_dep_imports.apps)
    libs_imports = flatten(int_dep_imports.libs)
    if not apps_imports and not libs_imports:
        return
    console = Console(theme=defaults.una_theme)
    table = Table(box=box.SIMPLE_HEAD)
    apps_i = int_dep_imports.apps
    libs_i = int_dep_imports.libs
    table.add_column("[data]int_dep[/]")
    table.add_column("[data]library[/]")
    for int_dep, imports in sorted(libs_i.items(), key=itemgetter(0)):
        table.add_row(f"[lib]{int_dep}[/]", ", ".join(sorted(imports)))
    for int_dep, imports in sorted(apps_i.items(), key=itemgetter(0)):
        table.add_row(f"[app]{int_dep}[/]", ", ".join(sorted(imports)))
    console.print(table, overflow="ellipsis")


def print_missing_installed_libs(
    int_dep_imports: OrgImports,
    third_party_libs: set[str],
    project_name: str,
) -> bool:
    style = config.get_style()
    include_libs = style == Style.modules
    diff = calculate_diff(int_dep_imports, third_party_libs, include_libs)
    if not diff:
        return True
    console = Console(theme=defaults.una_theme)
    missing = ", ".join(sorted(diff))
    console.print(
        f"[data]Could not locate all libraries in [/][proj]{project_name}[/]. [data]Caused by missing dependencies?[/]"
    )
    console.print(f":thinking_face: {missing}")
    return False


def printable_version(version: str | None) -> str:
    if version is None:
        ver = "-"
    elif version == "":
        ver = "✔"
    else:
        ver = version
    return f"[data]{ver}[/]"


def libs_in_projects_table(
    projects: list[Proj],
    libraries: set[str],
    options: Options,
) -> Table:
    table = Table(box=box.SIMPLE_HEAD)
    projects = sorted(projects, key=lambda p: p.name)
    project_headers = [f"[proj]{p.name}[/]" for p in projects]
    headers = ["[data]library[/]"] + project_headers
    for header in headers:
        table.add_column(header)
    for lib in sorted(libraries):
        proj_versions = [p.ext_deps.items.get(lib) for p in projects]
        printable_proj_versions = [printable_version(v) for v in proj_versions]
        cols = [markup.escape(lib)] + printable_proj_versions
        table.add_row(*cols)
    return table


def flattened_lib_names(projects: list[Proj]) -> set[str]:
    return {k for proj in projects for k in proj.ext_deps.items}


def print_libs_in_projects(projects: list[Proj], options: Options) -> None:
    flattened = flattened_lib_names(projects)
    if not flattened:
        return
    table = libs_in_projects_table(projects, flattened, options)
    console = Console(theme=defaults.una_theme)
    console.print(Padding("[data]Libraries in projects[/]", (1, 0, 0, 0)))
    console.print(table, overflow="ellipsis")



from pathlib import Path

from una import differ

root = Path.cwd()
ns = "my_namespace"
changed_files = [
    Path(f"libs/{ns}/a/core.py"),
    Path(f"some/other/{ns}/file.py"),
    Path(f"apps/{ns}/b/core.py"),
    Path(f"libs/{ns}/b/core.py"),
    Path(f"libs/{ns}/c/nested/subfolder/core.py"),
]


def test_get_changed_libs():
    res = differ.get_changed_libs(root, changed_files, ns)
    assert res == ["a", "b", "c"]


def test_get_changed_apps():
    res = differ.get_changed_apps(root, changed_files, ns)
    assert res == ["b"]



from rich.theme import Theme

una_theme = Theme(
    {
        "data": "#999966",
        "proj": "#8A2BE2",
        "lib": "#32CD32",
        "app": "#6495ED",
    }
)

keep_file_name = ".keep"
lock_file = "requirements.lock"
root_file = ".git"
pyproj = "pyproject.toml"

libs_dir = "libs"
apps_dir = "apps"
proj_dir = "projects"

packages_pyproj = """
[project]
name = "{name}"
version = "0.1.0"
description = ""
authors = []
dependencies = []
requires-python = "{python_version}"

[build-system]
requires = ["hatchling", "hatch-una"]
build-backend = "hatchling.build"

[tool.rye]
managed = true
dev-dependencies = []

[tool.hatch.build.hooks.una-build]
[tool.hatch.metadata.hooks.una-meta]

[tool.una.libs]
"""

test_template = """\
from {ns} import {name}
def test_import():
    assert {name}
"""



from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass, field
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Literal, TypeAlias, TypeVar

import tomlkit
from dataclasses_json import dataclass_json

Json: TypeAlias = dict[str, "Json"] | list["Json"] | str | int | float | bool | None
Imports: TypeAlias = dict[str, set[str]]
DepKind: TypeAlias = Literal["app", "lib"]


@dataclass
class OrgImports:
    apps: Imports = field(default_factory=dict)
    libs: Imports = field(default_factory=dict)


@dataclass
class Include:
    src: str
    dst: str
    # path: str
    # include: str
    # root: str


@dataclass(frozen=True)
class CheckReport:
    int_dep_imports: OrgImports
    ext_dep_imports: OrgImports
    int_dep_diff: set[str]
    ext_dep_diff: set[str]


@dataclass(frozen=True)
class Diff:
    name: str
    path: Path
    apps: set[str]
    libs: set[str]
    int_dep_imports: OrgImports


@dataclass
class Options:
    quiet: bool = False
    verbose: bool = False
    alias: list[str] = field(default_factory=list)


def rename_keys(old: str, new: str) -> Callable[[Json], None]:
    def rename_hyphens(d: Json) -> None:
        if isinstance(d, dict):
            for k, v in list(d.items()):
                rename_hyphens(v)
                if old in k:
                    d[k.replace(old, new)] = d.pop(k)

    return rename_hyphens


@dataclass_json
@dataclass(frozen=True)
class Project:
    name: str
    dependencies: list[str]
    version: str | None = None
    optional_dependencies: dict[str, list[str]] | None = None
    requires_python: str = ">= 3.8"


class Style(Enum):
    packages = "packages"
    modules = "modules"


@dataclass_json
@dataclass(frozen=True)
class Una:
    style: Style = Style.packages
    libs: dict[str, str] = field(default_factory=dict)


@dataclass_json
@dataclass(frozen=False)
class Tool:
    una: Una = field(default_factory=Una)


Self = TypeVar("Self", bound="Conf")


@dataclass_json
@dataclass()
class Conf:
    """
    Conf object.

    Should never be created manually, only loaded from a toml file.
    See the caveats on `to_str()`.
    """

    project: Project
    tool: Tool
    _tomldoc: tomlkit.TOMLDocument | None = field(default=None)

    if TYPE_CHECKING:
        # these are just here becaue dataclass_json doesn't
        # seem to play well with pyright?
        @classmethod
        def from_dict(cls: type[Self], _: Json) -> Self:
            raise

    @classmethod
    def from_tomldoc(cls: type[Self], tomldoc: tomlkit.TOMLDocument) -> Self:
        orig = deepcopy(tomldoc)
        rename_keys("-", "_")(tomldoc)
        res = cls.from_dict(tomldoc)
        res._tomldoc = orig
        return res

    def to_tomldoc(self) -> tomlkit.TOMLDocument:
        tomldoc = self._tomldoc
        if not tomldoc:
            raise ValueError("This Conf has no _tomldoc member. This should not happen")

        # impossible for project.dependencies to be unset as validated on load
        orig_deps: list[str] = tomldoc["project"]["dependencies"]  # type: ignore[reportIndexIssues]
        new_deps = set(self.project.dependencies) - set(orig_deps)
        for dep in new_deps:
            tomldoc["project"]["dependencies"].add_line(dep)  # type: ignore[reportIndexIssues]

        # deal with a a non-existent tool.una.libs
        try:
            tomldoc["tool"]["una"]["libs"].update(self.tool.una.libs)  # type: ignore[reportIndexIssues]
        except KeyError:
            una = tomlkit.table(True)
            libs = tomlkit.table()
            libs.update(self.tool.una.libs)  # type: ignore[reportUnknownMemberType]
            una.append("libs", libs)
            tomldoc["tool"].append("una", una)  # type: ignore[reportIndexIssues]
        return tomldoc

    @classmethod
    def from_str(cls: type[Self], s: str) -> Self:
        tomldoc = tomlkit.loads(s)
        return cls.from_tomldoc(tomldoc)

    def to_str(self) -> str:
        """
        Dump the config to a string.

        To preserve the original formatting and make my life easy, this function
        will currently only modify the following fields:
        - project.dependencies
        - tool.una.libs
        - tool.hatch.build.hooks.una-build
        - tool.hatch.meta.hooks.una-meta

        All others will be written from the original toml file.
        """
        tomldoc = self.to_tomldoc()
        return tomlkit.dumps(tomldoc)  # type: ignore[reportUnknownMemberType]


@dataclass(frozen=True)
class ConfWrapper:
    conf: Conf
    path: Path


@dataclass(frozen=True)
class ExtDeps:
    source: str
    items: dict[str, str]


@dataclass(frozen=True)
class IntDeps:
    libs: list[str] = field(default_factory=list)
    apps: list[str] = field(default_factory=list)


@dataclass(frozen=False)
class Proj:
    name: str
    packages: list[Include]
    path: Path
    ext_deps: ExtDeps
    int_deps: IntDeps = field(default_factory=IntDeps)



from una import alias


def test_parse_one_key_one_value_alias():
    res = alias.parse(["opencv-python=cv2"])
    assert res["opencv-python"] == ["cv2"]
    assert len(res.keys()) == 1


def test_parse_one_key_many_values_alias():
    res = alias.parse(["matplotlib=matplotlib, mpl_toolkits"])
    assert res["matplotlib"] == ["matplotlib", "mpl_toolkits"]
    assert len(res.keys()) == 1


def test_parse_many_keys_many_values_alias():
    res = alias.parse(["matplotlib=matplotlib, mpl_toolkits", "opencv-python=cv2"])
    assert res["matplotlib"] == ["matplotlib", "mpl_toolkits"]
    assert res["opencv-python"] == ["cv2"]
    assert len(res.keys()) == 2


def test_pick_alias_by_key():
    aliases = {"opencv-python": ["cv2"]}
    keys = {"one", "two", "opencv-python", "three"}
    res = alias.pick(aliases, keys)
    assert res == {"cv2"}


def test_pick_aliases_by_keys():
    aliases = {"opencv-python": ["cv2"], "matplotlib": ["mpl_toolkits", "matplotlib"]}
    keys = {"one", "two", "opencv-python", "matplotlib", "three"}
    res = alias.pick(aliases, keys)
    assert res == {"cv2", "mpl_toolkits", "matplotlib"}


def test_pick_empty_alias_by_keys():
    aliases: dict[str, list[str]] = {}
    keys = {"one", "two", "opencv-python", "matplotlib", "three"}
    res = alias.pick(aliases, keys)
    assert res == set()



import ast
from functools import lru_cache
from pathlib import Path

from una.types import Imports


def parse_import(node: ast.Import) -> list[str | None]:
    return [name.name for name in node.names]


def extract_import_from(node: ast.ImportFrom) -> list[str | None]:
    return [f"{node.module}.{alias.name}" for alias in node.names] if node.names else [node.module]


def parse_import_from(node: ast.ImportFrom) -> list[str | None]:
    return extract_import_from(node) if node.module and node.level == 0 else []


def parse_imports(node: ast.AST) -> list[str | None]:
    if isinstance(node, ast.Import):
        return parse_import(node)
    if isinstance(node, ast.ImportFrom):
        return parse_import_from(node)
    return []


@lru_cache
def parse_module(path: Path) -> ast.AST:
    with open(path.as_posix(), encoding="utf-8", errors="ignore") as f:
        tree = ast.parse(f.read(), path.name)
    return tree


def extract_imports(path: Path) -> list[str]:
    tree = parse_module(path)
    return [i for node in ast.walk(tree) for i in parse_imports(node) if i is not None]


def list_imports(path: Path) -> set[str]:
    py_modules = path.rglob("*.py")
    extracted = (extract_imports(m) for m in py_modules)
    flattened = (i for imports in extracted for i in imports)
    return set(flattened)


def fetch_all_imports(paths: set[Path]) -> Imports:
    rows = [{p.name: list_imports(p)} for p in paths]
    return {k: v for row in rows for k, v in row.items()}



from una import internal_deps

apps_in_ws = {"x"}
libs_in_ws = {"a", "b"}


def test_to_row_returns_columns_for_all_int_deps():
    expected_length = len(apps_in_ws) + len(libs_in_ws)
    collected_import_data = {"x": {"a"}, "a": {"b"}}
    flattened_imports: set[str] = set().union(*collected_import_data.values())
    rows = internal_deps.create_rows(
        apps_in_ws,
        libs_in_ws,
        collected_import_data,
        flattened_imports,  # type: ignore[reportArgumentType]
    )
    assert len(rows) == expected_length
    for columns in rows:
        assert len(columns) == expected_length



from collections.abc import Sequence
from pathlib import Path

from rich.console import Console

from una import check, config, defaults, files, internal_deps
from una.config import load_conf
from una.types import Conf, Diff, Include, Options, Proj, Style


def sync_project_int_deps(root: Path, ns: str, project: Proj, options: Options):
    apps_pkgs = files.get_apps(root, ns)
    libs_pkgs = files.get_libs(root, ns)
    diff = calculate_diff(root, ns, project, apps_pkgs, libs_pkgs)
    update_project(ns, diff)
    if options.quiet:
        return
    print_summary(diff)
    if options.verbose:
        print_int_dep_imports(diff)


def calculate_diff(
    root: Path,
    ns: str,
    project: Proj,
    apps_pkgs: Sequence[str],
    libs_pkgs: Sequence[str],
) -> Diff:
    proj_apps = set(project.int_deps.apps)
    proj_libs = set(project.int_deps.libs)
    all_apps = set(apps_pkgs)
    all_libs = set(libs_pkgs)
    int_dep_imports = internal_deps.get_int_dep_imports(root, ns, proj_apps, proj_libs)
    int_dep_diff = check.imports_diff(int_dep_imports, proj_apps, proj_libs)
    apps_diff = {b for b in int_dep_diff if b in all_apps}
    libs_diff = {b for b in int_dep_diff if b in all_libs}
    return Diff(
        name=project.name,
        path=project.path,
        apps=apps_diff,
        libs=libs_diff,
        int_dep_imports=int_dep_imports,
    )


def print_int_dep_imports(diff: Diff) -> None:
    int_dep_imports = diff.int_dep_imports
    check.print_int_dep_imports(int_dep_imports)


def print_summary(diff: Diff) -> None:
    console = Console(theme=defaults.una_theme)
    name = diff.name
    apps_pkgs = diff.apps
    libs_pkgs = diff.libs
    anything_to_sync = apps_pkgs or libs_pkgs
    emoji = ":point_right:" if anything_to_sync else ":heavy_check_mark:"
    printable_name = f"[proj]{name}[/]"
    console.print(f"{emoji} {printable_name}")
    for b in apps_pkgs:
        console.print(f"adding [app]{b}[/] app to [proj]{name}[/]")
    for c in libs_pkgs:
        console.print(f"adding [lib]{c}[/] lib to [proj]{name}[/]")
    if anything_to_sync:
        console.print("")


def to_package(ns: str, name: str, int_dep_root: str, style: Style) -> Include:
    root = Path(int_dep_root)
    src = root / ns / name if style == Style.modules else root / name / ns / name
    dst = Path(ns) / name
    return Include(src=str(src), dst=str(dst))


def generate_updated_project(conf: Conf, packages: list[Include]) -> str | None:
    for inc in packages:
        conf.tool.una.libs[inc.src] = inc.dst
    return conf.to_str()


def to_packages(ns: str, diff: Diff) -> list[Include]:
    style = config.get_style()
    apps_path = "../../apps"
    libs_path = "../../libs"
    a = [to_package(ns, b, apps_path, style) for b in diff.apps]
    b = [to_package(ns, c, libs_path, style) for c in diff.libs]
    return a + b


def rewrite_project_file(path: Path, packages: list[Include]):
    conf = load_conf(path)
    generated = generate_updated_project(conf, packages)
    if not generated:
        return
    fullpath = path / defaults.pyproj
    with fullpath.open("w", encoding="utf-8") as f:
        f.write(generated)


def update_project(ns: str, diff: Diff):
    packages = to_packages(ns, diff)
    if packages:
        rewrite_project_file(diff.path, packages)



import re
import subprocess
from pathlib import Path

from rich.console import Console
from rich.padding import Padding

from una import config, defaults, internal_deps
from una.types import Proj


def calc_diff(tag_name: str | None, only_int_deps: bool):
    root = config.get_workspace_root()
    tag = get_latest_tag(tag_name)
    if not tag:
        print("No tags found in repository.")
        return

    ns = config.get_ns()
    files = get_files(tag)
    apps_paths = get_changed_apps(root, files, ns)
    libs_paths = get_changed_libs(root, files, ns)
    projects = get_changed_projects(files)
    all_projects_data = internal_deps.get_int_deps_in_projects(root, libs_paths, apps_paths, ns)
    projects_data = [p for p in all_projects_data]
    if only_int_deps:
        print_detected_changes_in_int_deps(apps_paths, libs_paths)
        return
    print_diff_summary(tag, apps_paths, libs_paths)
    print_detected_changes(projects, "proj")
    print_diff_details(projects_data, apps_paths, libs_paths)


def parse_folder_parts(pattern: str, changed_file: Path) -> str:
    parts = re.split(pattern, changed_file.as_posix())
    remainder = parts[-1]
    file_path = Path(remainder)
    return next(p for p in file_path.parts if p != file_path.root)


def get_changed(pattern: str, changed_files: list[Path]) -> set[str]:
    return {parse_folder_parts(pattern, f) for f in changed_files if re.match(pattern, f.as_posix())}


def parse_path_pattern(_: Path, top_dir: str, namespace: str) -> str:
    return f"{top_dir}/{namespace}/"


def get_changed_int_deps(root: Path, top_dir: str, changed_files: list[Path], namespace: str) -> list[str]:
    pattern = parse_path_pattern(root, top_dir, namespace)
    return sorted(get_changed(pattern, changed_files))


def get_changed_libs(root: Path, changed_files: list[Path], namespace: str) -> list[str]:
    return get_changed_int_deps(root, defaults.libs_dir, changed_files, namespace)


def get_changed_apps(root: Path, changed_files: list[Path], namespace: str) -> list[str]:
    return get_changed_int_deps(root, defaults.apps_dir, changed_files, namespace)


def get_changed_projects(changed_files: list[Path]) -> list[str]:
    res = get_changed(defaults.apps_dir, changed_files)
    filtered = {p for p in res if p != defaults.apps_dir}
    return sorted(filtered)


def get_latest_tag(key: str | None) -> str | None:
    tag_pattern = config.get_tag_pattern(key)
    sorting_options = [f"--sort={option}" for option in config.get_tag_sort_options()]
    res = subprocess.run(
        ["git", "tag", "-l"] + sorting_options + [f"{tag_pattern}"],
        capture_output=True,
    )
    return next((tag for tag in res.stdout.decode("utf-8").split()), None)


def get_files(tag: str) -> list[Path]:
    res = subprocess.run(
        ["git", "diff", tag, "--stat", "--name-only"],
        capture_output=True,
    )
    return [Path(p) for p in res.stdout.decode("utf-8").split()]


def print_diff_details(projects_data: list[Proj], apps: list[str], libs: list[str]) -> None:
    if not apps and not libs:
        return
    console = Console(theme=defaults.una_theme)
    table = internal_deps.build_int_deps_in_projects_table(projects_data, apps, libs, for_info=False)
    console.print(table, overflow="ellipsis")


def print_detected_changes(changes: list[str], markup: str) -> None:
    if not changes:
        return
    console = Console(theme=defaults.una_theme)
    for int_dep in changes:
        console.print(f"[data]:gear: Changes found in [/][{markup}]{int_dep}[/]")


def print_detected_changes_in_int_deps(apps: list[str], libs: list[str]) -> None:
    sorted_apps = sorted(apps)
    sorted_libs = sorted(libs)
    print_detected_changes(sorted_libs, "lib")
    print_detected_changes(sorted_apps, "app")


def print_diff_summary(tag: str, apps: list[str], libs: list[str]) -> None:
    console = Console(theme=defaults.una_theme)
    console.print(Padding(f"[data]Diff: based on the {tag} tag[/]", (1, 0, 1, 0)))
    if not apps and not libs:
        console.print("[data]No int_dep changes found.[/]")
        return
    if libs:
        console.print(f"[lib]Changed libs[/]: [data]{len(libs)}[/]")
    if apps:
        console.print(f"[app]Changed apps[/]: [data]{len(apps)}[/]")
